#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <malloc.h>
#include <linux/cdrom.h>
#include <linux/optcd.h>

#define CDNULL 0
#define CDPLAY 1
#define CDPAUZ 3
#define CDSTOP 4
#define CDEJECT 5

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

struct CDTrack{
   int length;		/* Length of track in seconds or Kbytes */
   int start;		/* Starting position (f+s*75+m*60*75) */
   int track;		/* Physical track number */
   char data;		/* Flag: data track */
};

struct CDDisc{
   int ntracks;                 /* Number of tracks on the disc */
   int length;                  /* Total running time in seconds */
   struct CDTrack *tracks;	/* struct trackinfo[ntracks] */
};

struct CDStat{
   int track;
   int index;
   int pos_abs;
   int frame;
   int pos_rel;
   int tracklen;
   int remain;
   int cdlen;
   int ntracks;	
   int cdmode;
   bool datatrk;
   bool trayopen;
};

class CDCtl
{
public:
   CDCtl(char *dname){
      device=dname;
      cdstat=NULL;
      cddisc=NULL;
      if((cdfd=open(device,0))==-1)
         cdfdopen=FALSE;
      else{
         cdfdopen=TRUE;
         cdstat=(struct CDStat *)malloc(sizeof(struct CDStat));
         cdstat->cdmode=5;
	 cdstat->track = 1;
         cdstat->trayopen=FALSE;
         status();
      }
   }
   ~CDCtl(){
      if(cdfdopen)
         close(cdfd);
      if(cddisc!=NULL){
         if(cddisc->tracks!=NULL)
            free(cddisc->tracks);
         free(cddisc);
      }
      if(cdstat!=NULL)
         free(cdstat);
   }
   void btnEject(){
      if(cdstat!=NULL){
	 if(cdstat->trayopen)
            closetray();
	 else{
            stop();
            eject();
	 }
         status();
      }
   }
   void btnStop(){
      stop();
      status();
   }
   void btnPause(){
      if(cdstat!=NULL){
         if(cdstat->cdmode==3)
            resume();
         else
            pause();
         status();
      }
   }
   void btnPlay(){
      if(cdstat!=NULL){
         play(cdstat->track-1);
         status();
      }
   }
   void btnPrev(){
      if(cdstat!=NULL){
         cdstat->track--;
         if(cdstat->track<1)
            cdstat->track=cdstat->ntracks;
         if(cdstat->cdmode==1)
           play(cdstat->track-1);
         status();
      }
   }
   void btnNext(){
      if(cdstat!=NULL){
         cdstat->track++;
         if(cdstat->track>cdstat->ntracks)
            cdstat->track=1;
         if(cdstat->cdmode==1)
           play(cdstat->track-1);
         status();
      }
   }
   void btnFfwd(){
      if(cdstat!=NULL){
         int start=cdstat->frame+15*75;
         int end=cdstat->cdlen*75;
         if(start>end)
            start-=end;
         play_msf(start,end);
         status();
      }
   }
   void btnRewd(){
      if(cdstat!=NULL){
         int start=cdstat->frame-15*75;
         int end=cdstat->cdlen*75;
         if(start<0)
            start+=end;
         play_msf(start,end);
         status();
      }
   }
   void eject(){
      if(cdfdopen && cdstat!=NULL)
         cdstat->trayopen=!ioctl(cdfd,CDROMEJECT);
   }
   void closetray(){
      if(cdfdopen && cdstat!=NULL)
         cdstat->trayopen=!ioctl(cdfd,CDROMCLOSETRAY);
   }
   void pause(){
      if(cdfdopen)
         ioctl(cdfd,CDROMPAUSE);
   }
   void resume(){
      if(cdfdopen)
         ioctl(cdfd,CDROMRESUME);
   }
   void stop(){
      if(cdfdopen)
         ioctl(cdfd,CDROMSTOP);
   }
   void play(int trk){
      if(cddisc!=NULL){
	 if(cddisc->tracks[trk].data)
            stop();
         else{
            int start=cddisc->tracks[trk].start;
            int end=(cddisc->length*75);
            play_msf(start,end);
         }
      }
   }
   void status(){
      if(cdfdopen){
         struct cdrom_subchnl sc;
         sc.cdsc_format=CDROM_MSF;
         if(ioctl(cdfd, CDROMSUBCHNL, &sc)){
            cdstat->cdmode = 5;
	    cdstat->track = 1;
         }
         else{
            cdstat->trayopen=FALSE;
	    if(cdstat->cdmode==5){
	       readTOC();
               if(cddisc!=NULL){
                  cdstat->cdmode=4;
		  cdstat->ntracks=cddisc->ntracks;
                  cdstat->cdlen=cddisc->length;
                  trackinfo(&sc);
	       }
            }
            else
               cdstat->cdmode=0;
            if(cddisc!=NULL){
	       switch(sc.cdsc_audiostatus){
	        case CDROM_AUDIO_PLAY:
                   cdstat->cdmode=1;
                break;
	        case CDROM_AUDIO_PAUSED:
                   cdstat->cdmode=3;
                break;
                case CDROM_AUDIO_COMPLETED:
                   cdstat->cdmode=0;
                break;
	        case CDROM_AUDIO_NO_STATUS:
                   cdstat->cdmode=4;
                break;
	       }
            }
            trackinfo(&sc);
         }
      }
   }
   void readTOC(){
      if(cddisc==NULL)
         cddisc=(struct CDDisc *)malloc(sizeof(struct CDDisc));
      else{
         if(cddisc->tracks!=NULL)
            free(cddisc->tracks);
      }
      if(cddisc!=NULL){
         struct cdrom_tochdr hdr;
         if(ioctl(cdfd, CDROMREADTOCHDR, &hdr)){
            free(cddisc);
            cddisc=NULL;
         }
         else{
            cddisc->length=0;
            cddisc->ntracks=hdr.cdth_trk1;
            cddisc->tracks=(struct CDTrack *)malloc(cddisc->ntracks*sizeof(struct CDTrack));
            if(cddisc->tracks==NULL){
               free(cddisc);
               cddisc=NULL;
            }
            else{
               struct cdrom_tocentry te;
               int prev_addr=0;
               for(int i=0;i<=cddisc->ntracks;i++){
                  if(i==cddisc->ntracks)
                     te.cdte_track=CDROM_LEADOUT;
                  else
                     te.cdte_track=i+1;
                  te.cdte_format=CDROM_MSF;
                  if(ioctl(cdfd, CDROMREADTOCENTRY, &te)){
                     free(cddisc->tracks);
                     free(cddisc);
                     cddisc=NULL;
                     break;
                  }
                  else{
                     if(i>0)
		        cddisc->tracks[i-1].length = te.cdte_addr.msf.minute*60 + te.cdte_addr.msf.second - prev_addr;
                     prev_addr=te.cdte_addr.msf.minute*60 + te.cdte_addr.msf.second;
                     if(i<cddisc->ntracks){
                        cddisc->tracks[i].data = te.cdte_ctrl & CDROM_DATA_TRACK ? 1 : 0;
	                cddisc->tracks[i].start = (te.cdte_addr.msf.minute*60 + te.cdte_addr.msf.second)*75  + te.cdte_addr.msf.frame;
		        cddisc->tracks[i].track = i + 1;
                     }
                     else
                        cddisc->length = te.cdte_addr.msf.minute*60 + te.cdte_addr.msf.second;
                  }
               }
            }
	 }
      }
   }
   void trackinfo(struct cdrom_subchnl *subchnl){
      cdstat->index=subchnl->cdsc_ind;
      cdstat->pos_abs=subchnl->cdsc_absaddr.msf.minute*60 + subchnl->cdsc_absaddr.msf.second;
      cdstat->frame=cdstat->pos_abs*75 + subchnl->cdsc_absaddr.msf.frame;
      if(subchnl->cdsc_trk!=0 && (cdstat->cdmode==1 || cdstat->cdmode==3)){
         cdstat->track=subchnl->cdsc_trk;
         //track correction (i don't know why but it seems to advance the track early sometimes)
         if(cdstat->frame < cddisc->tracks[cdstat->track-1].start && cdstat->frame >= cddisc->tracks[0].start)
            cdstat->track--;
      }
      cdstat->pos_rel=(cdstat->frame - cddisc->tracks[cdstat->track-1].start)/75;
      if(cdstat->track<cdstat->ntracks)
         cdstat->remain=cddisc->tracks[cdstat->track].start-cdstat->frame;
      else
         cdstat->remain=cdstat->cdlen*75-cdstat->frame;
      cdstat->tracklen=cddisc->tracks[cdstat->track-1].length;
      cdstat->datatrk=cddisc->tracks[cdstat->track-1].data;
   }
   struct CDDisc *cddisc;
   struct CDStat *cdstat;
private:
   void play_msf(int start, int end){
      if(cdfdopen){
         struct cdrom_msf playmsf;
	 playmsf.cdmsf_min0 = start / (60*75);
         playmsf.cdmsf_sec0 = (start % (60*75)) / 75;
	 playmsf.cdmsf_frame0 = start % 75;
	 playmsf.cdmsf_min1 = end / (60*75);
	 playmsf.cdmsf_sec1 = (end % (60*75)) / 75;
	 playmsf.cdmsf_frame1 = end % 75;
         ioctl(cdfd,CDROMPLAYMSF,&playmsf);
      }
   }
   int cdfd;
   int cdfdopen;
   char *device;
};
